(* Simple CGI interface *)


(* CGI environment *)

module HEnv = Map.Make(String);;
type hvar = string
type henv = string HEnv.t

let hlocate key henv = try Some (HEnv.find key henv) with Not_found -> None

(* Data type of values being exchanged with a Web client *)
type u = UI of int | UF of float | US of string;;
let ushow = function
  | UI x -> string_of_int x
  | UF x -> string_of_float x
  | US x -> x
;;

type ('a,'b) either = Left of 'a | Right of 'b;;

(* Input validation routines *)
type 'a validation_fn = string option -> (string,'a) either

let vstring = function 
  | None | Some ""  -> Left "value is not given"
  | Some s -> Right s
let vint = function 
  | None | Some ""  -> Left "value is not given"
  | Some s -> (try Right (int_of_string s)
               with _ -> Left ("bad integer number: " ^ s))
let vfloat = function
  | None | Some "" -> Left "value is not given"
  | Some s -> (try Right (float_of_string s)
               with _ -> Left ("bad floating-point number: " ^ s))
let vbool = function
  | Some _ -> Right true
  | None   -> Right false

(* Given henv, the name of the expected variable and the validation
   function, extract the desired input and validate it.
   Invoke inquirer to report an error and ask for a corrected reply.
*)
let rec validate inquirer hvar henv validfn =
  match validfn (hlocate hvar henv) with
    | Left err -> 
	let henv = 
	  HEnv.add "response" ("Bad field `" ^ hvar ^ "': " ^ err) henv in
	  validate inquirer hvar (inquirer henv) validfn
    | Right v  -> v


(*
  HTML encoding
  We are very strict: we permit only ASCII printable characters,
  and CR/LF/TAB
  Everything else is replaced with #
  In particular, no Unicode other than ASCII is allowed!
*)
let html_encode_char b : char -> unit = function
  | (' ' | '\n' | '\t' as c) -> Buffer.add_char b c
  | '\r' -> ()
  | '<'  -> Buffer.add_string b "&lt;"
  | '>'  -> Buffer.add_string b "&gt;"
  | '"'  -> Buffer.add_string b "&quot;"
  | '&'  -> Buffer.add_string b "&amp;"
  | (' '..'~' as c) -> Buffer.add_char b c
  | _ -> Buffer.add_char b '#'
;;

let html_encode_string str : string =
  let b = Buffer.create (String.length str) in
  String.iter (html_encode_char b) str;
  Buffer.contents b
;;



(*
  Parse the string of URL parameters, the "QUERY_STRING"
  That is, convert a string like
 	"parm1=val1&parm2=val2+val%253+val4&%7Eparm3=&parm4"
  into an assoc list
 	'((parm4) (~parm3 "")
          (parm2 "val2" "val%3" "val4")
          (parm1 "val1"))
 
  Parsing is done by a finite state machine, that takes into
  account the current state (looking for a parm, looking for a value,
  looking for a continuation of parm/value after %xx quoted-char),
  the action-prefix ('=', '+', '&' or '/') and the token that follows
  the action prefix (that is, a set of characters through the next
  action-prefix or the end-of-string). At the very beginning, the
  action-prefix is assumed '&'. Note, we assume '/' separator to
  act exactly like '&' (so we can use the present function to
  parse both SCRIPT_PATH and QUERY_STRINGs)
*)

(* The corresponding Haskell code is far more elegant because it is
   purely functional and strings are streams in Haskell.
   Unfortunately, OCaml's streams, buffers are all very imperative.
*)
let url_unquote str = 
 let token = Buffer.create 16 in (* accumulate the current token *)
 let whole_token () = let k  = Buffer.contents token in
                      let () = Buffer.clear token in k in
 let stream = Stream.of_string str in
 let next () = try Some (Stream.next stream) with Stream.Failure -> None in
 let unq =     (* Here, we prohibit anything above 126 and 0 *)
   let build n = if n > 126 || n < 10 then "*forbidden-percent*"
                 else String.make 1 (Char.chr n) in
   function
   | (Some c1, Some c2) -> 
       (try
	 Scanf.sscanf ((String.make 1 c1) ^ (String.make 1 c2)) "%x" build
        with Scanf.Scan_failure _ -> "*bad-percent*")
   | _ -> "*underflow-percent*"
 in
 (* searching for the key, accumulate in token *)
 let rec key henv =
   match next () with
   | None -> let k = whole_token () in (* Finished *)
             if String.length k = 0 then henv else (k,"")::henv
   | Some '&' ->	(* key with no value *)
       key ((whole_token (),"")::henv)
   | Some '=' ->  v_l (whole_token ()) henv
   | Some '%' ->
       let c1 = next () in let c2 = next () in
       let () = Buffer.add_string token (unq (c1,c2)) in
       key henv
   | Some '+' -> let () = Buffer.add_char token ' ' in key henv
   | Some x   -> let () = Buffer.add_char token x   in key henv
 and v_l k henv =     (* we have read the key; now read the value *)
   match next () with
   | None -> (k,whole_token ())::henv
   | Some '&' -> key ((k,whole_token ())::henv)
   | Some '=' -> key ((k,whole_token ())::henv) (* repeated =, weird. assume &*)
   | Some '%' ->
       let c1 = next () in let c2 = next () in
       let () = Buffer.add_string token (unq (c1,c2)) in
       v_l k henv
   | Some '+' -> let () = Buffer.add_char token ' ' in v_l k henv
   | Some x   -> let () = Buffer.add_char token x   in v_l k henv
 in key []
;;

let test_unq () = List.map url_unquote
	       ["aaa";
		"aaa=";
		"aaa=%%";
		"aaa=%ff";
		"aaa=&bbb=b1&ccc=c1+c2&ddd";
		"%7eaaa=&%25b%25bb=b%201&c%20c%7E=c1+c2&ddd%21";
	        "parm=+val1++val2+&parm2"]
;;
(*
[[("aaa", "")]; 
 [("aaa", "")]; 
 [("aaa", "*underflow-percent*")];
 [("aaa", "*forbidden-percent*")];
 [("ddd", ""); ("ccc", "c1 c2"); ("bbb", "b1"); ("aaa", "")];
 [("ddd!", ""); ("c c~", "c1 c2"); ("%b%bb", "b 1"); ("~aaa", "")];
 [("parm2", ""); ("parm", " val1  val2 ")]]
*)

(* Just like url_unquote, but merge the entries with the same keys.
   This may happen (e.g., multiple-choice radio boxes)
*)

let url_unquote_collate str =
  List.fold_left (fun m (k,v) ->
    HEnv.add k (try (HEnv.find k m) ^ "/" ^ v with Not_found -> v) m)
    HEnv.empty (url_unquote str)
;;

let test_unq' () = 
  List.map (fun s -> 
    HEnv.fold (fun k v l -> (k,v)::l) (url_unquote_collate s) [])
	       ["aaa";
		"aaa=";
		"k1=1&k2=2&k1=3"]
;;
(*
 - : (HEnv.key * string) list list =
 [[("aaa", "")]; [("aaa", "")]; [("k2", "2"); ("k1", "3/1")]]
*)

(* A base error handler, to be used when none of the HTTP headers are sent. 
   It should send all the headers and 'reset' (because this is a CGI script, 
   reset means finish)
*)
let base_error_handler e =
    Printf.eprintf "base_error_handler: %s\n" e;
    print_string (String.concat "\n"
      [
       "Content-type: text/html";
       "";
       "<html><head><title>CGI script error</title></head>";
       "<body><h1>CGI script error</h1>";
       "<p><b>"; html_encode_string e; "</b>";
       "</body></html>"]);
    flush_all ();
    exit 4
;;

(*
  Try to extract the form environment from the process environment,
  return HEnv.empty if  we haven't been called by a CGI script
*)
let get_form_env () = 
  let get_query' = function
    | "GET" -> Sys.getenv "QUERY_STRING"
    | "POST" -> 
	let len = int_of_string (Sys.getenv "CONTENT_LENGTH") in
	let ctype = Sys.getenv "CONTENT_TYPE" in
	if String.lowercase ctype = "application/x-www-form-urlencoded" then
	  let query = String.create len in
	  really_input stdin query 0 len; query
	else failwith ("bad ctype for POST: " ^ ctype)
    | rm -> failwith ("bad request method: " ^ rm)
  in
  let get_query = function 
    | None -> HEnv.empty
    | Some req_method ->
	let self = Sys.getenv "SCRIPT_NAME" in
	let henv = url_unquote_collate (get_query' req_method) in
	HEnv.add "this-script" self
	  (HEnv.add "request-method" req_method henv)
  in
  get_query (try Some (Sys.getenv "REQUEST_METHOD") with Not_found -> None)
;;


(* Write the form to the standard output, substituting the variables
   according to HEnv
   The form may include `variables'
   ${varname}
   where varname is a token
   If ${varname} appears within a form field, please make sure
   the name of that variable and the name of the form field are the same

   Errors are allowed to propagate to the top level and kill the process
   The output sent to the browser will probably be truncated.
   Oh well, imagine if there was a network error: the HTML received
   by the browser will be truncated too. The browser has to deal with this
   anyway.
*)

let send_form cout henv_raw form_name = 
  let henv = HEnv.map html_encode_string henv_raw in
  let token = Buffer.create 16 in
  let cin = open_in form_name in
  let next () = try Some (input_char cin) with End_of_file -> None in
  let rec loop = function
    | None -> ()
    | Some '$' -> loop1 (next ())
    | Some c -> output_char cout c; loop (next ())
  and loop1 = function 			(* seen $ *)
    | None -> output_char cout '$'
    | Some '{' -> loop2 (next ())
    | Some x -> output_char cout '$'; output_char cout x; loop (next ())
  and loop2 = function			(* seen ${ *)
    | None -> failwith "Non-terminated variable name in template"
    | Some '}' -> do_subst ()
    | Some ('a'..'z' | 'A'..'Z' | '_' | '0'..'9' | '-' as c) ->
	Buffer.add_char token c; loop2 (next ())
    | _ -> failwith "Bad character in a variable name in template"
  and do_subst = function () ->
    let k = Buffer.contents token in
    let () = Buffer.clear token in
    let () = try output_string cout (HEnv.find k henv) 
             with Not_found -> 
	       Printf.eprintf "Variable %s not found in %s" k form_name in
    loop (next ())
  in loop (next ()); close_in_noerr cin
;;

(* Various utilities *)

let maybe n j = function Some x -> j x | None -> n ()

let splitby str char =
  let s = Stream.of_string str in
  let b = Buffer.create 16 in
  let iseof s = maybe (fun () -> true) (fun _ -> false) (Stream.peek s) in
  let rec loop found =
    match Stream.peek s with
    | None -> List.rev (Buffer.contents b :: found)
    | Some c when c = char ->
	let _ = Stream.next s in
	let found = Buffer.contents b :: found in
	let () = Buffer.reset b in
	if iseof s then List.rev found else loop found
    | Some c -> let _ = Stream.next s in Buffer.add_char b c; loop found
  in loop []
;;

(*
# splitby "abc" '&';;
- : string list = ["abc"]
# splitby "abc&" '&';;
- : string list = ["abc"]
# splitby "abc&&" '&';;
- : string list = ["abc"; ""]
# splitby "&abc&&" '&';;
- : string list = [""; "abc"; ""]
*)
