(* Continuation-based CGI interface *)
(* This file uses persistent delimited continuations and so
   supports the regular CGI.
*)

open Cgi
open Delimcc
open Kcgi

type cgi_result' = Res of string * (henv,(hvar * string) list) either
type cgi_result = unit -> cgi_result'
type k_t = henv -> cgi_result


(* where cont are stored. Should be consistent with the directory
   used by Filename.temp_file
*)
let cont_dir = "/tmp"			


let save_state_file fname k =
  let oc = open_out fname in
  let () = output_delim_value oc k in
  close_out oc

let load_state_file fname =
  let ic = open_in fname in
  let v = Marshal.from_channel ic in
  let () = close_in ic in
  Printf.eprintf "Loaded k from the file %s\n" fname;
  v

let locate_cont klabel : k_t = 
  Obj.obj(absolutize (load_state_file (Filename.concat cont_dir klabel)))



let tx_label_name = "&TXLABEL&"

let in_transaction exn env body =
  let txprev = try Some (List.assoc tx_label_name env) with Not_found -> None in
  let txfile = Filename.temp_file "tx" "" in
  let () = close_out (open_out txfile) in (* create tx file *)
  let txname = match txprev with None -> txfile 
               | Some txprev -> txprev ^ "-" ^ txfile in
  let env    = (tx_label_name,txname)::env in
  let res = 
    try body env
    with e -> let () = try Sys.remove txfile with _ -> () in raise e
  in
  let () = try Sys.remove txfile with Sys_error e -> raise exn in
  res
;;

(* This is like writing followed by reading. We combine them
   for sake of input validation.
*)

let add k v = function
  | Left henv -> Left (HEnv.add k v henv)
  | Right env -> Right ((k,v)::env)

let find k = function
  | Left henv -> hlocate k henv
  | Right env -> try Some (List.assoc k env) with Not_found -> None

let check_transaction exn env =
  match find tx_label_name env with
  | None -> None
  | Some txname as r -> 
      Printf.eprintf "transaction check: %s\n" txname;
      List.iter 
	(fun fname -> if Sys.file_exists fname then () else raise exn)
	(splitby txname '-');
      r
      

let sendk p template env (k : k_t) () =
  let fname  = Filename.temp_file "kcgi" "" in
  let klabel = Filename.basename fname in
  let () = prerr_endline ("saving in " ^ fname) in
  (* let () = show_val 5 (Obj.repr k) in *)
  let () = save_state_file fname (Obj.repr k) in
  let () = prerr_endline ("cont saved in " ^ fname) in
  Res (template,add "klabel" klabel env)

let do_inquire exn p template env =
  Printf.eprintf "inquire': %s\n" template;
  (* prerr_endline "p in inquire'\n"; show_val 5 (Obj.repr p); *)
  let txname = check_transaction exn env in
  let henv = shift p (sendk p template env) in
  (*
  let () = prerr_endline "p after shift\n"; 
      ignore(relativitize (Obj.repr (true,p)) true);
      show_val 5 (Obj.repr p) in
  *)
  let henv = 
    match txname with None -> henv 
    | Some txn -> HEnv.add tx_label_name txn henv in
  HEnv.add hvar_template_name template henv


let answer' exn p hvar henv validfn = 
  validate (fun henv -> 
    do_inquire exn p (HEnv.find hvar_template_name henv) (Left henv))
    hvar henv validfn

let () =
  let () = init_global_closure (do_inquire,sendk,validate,
				HEnv.add,hlocate,answer',in_transaction,
			        Blog.read_blog, Blog.write_blog) in
  prerr_endline "Initialized\n\n"

(* End of global data *)

module KCGI_P(S:sig val p0 : cgi_result prompt end) = struct 
open S
exception TX_Gone

let () = prerr_endline "p0 at startup\n"; show_val 5 (Obj.repr p0)

let inquire template outputs =
  do_inquire TX_Gone p0 template (Right outputs)

let inquire_finish template outputs =
  abort p0 (fun () -> (Res (template,Right outputs)))

let answer hvar henv validfn = answer' TX_Gone p0 hvar henv validfn

let in_transaction body = in_transaction TX_Gone body
end

module Run(M:MAIN) = struct

 let our_p0 : cgi_result prompt = new_prompt ()
 module Kcgi = KCGI_P(struct let p0 = our_p0 end)
 module Main = M(Kcgi)
 open Main

let run () = 
  let (template,henv) =
    try
      let henv = get_form_env () in
      let Res (template,env) =
	match 
	  try Some (locate_cont (HEnv.find "klabel" henv))
	    with Not_found -> None | Sys_error _ -> None with
	  | None -> push_prompt our_p0 (fun () -> main (); failwith "na") ()
	  | Some k -> k henv () in
	let henv = match env with
	| Left henv' -> HEnv.fold HEnv.add henv' henv 
	| Right env  ->
	    List.fold_left (fun m (k,v) -> HEnv.add k v m) henv env in
	(template,henv)
      with e -> base_error_handler (Printexc.to_string e) in
    (* Now we are creating output *)
    print_endline "Content-type: text/html\n";
    send_form stdout henv (Filename.concat template_dir template);
    flush_all ();
    exit 0
end
